from dune.perftool import Restriction
from dune.perftool.generation import (preamble,
                                      symbol,
                                      temporary_variable,
                                      )
from dune.perftool.pdelab.quadrature import (name_quadrature_position,
                                             quadrature_preamble,
                                             )


@symbol
def name_element_geometry_wrapper():
    return 'eg'


@symbol
def type_element_geometry_wrapper():
    return 'EG'


@symbol
def name_intersection_geometry_wrapper():
    return 'ig'


@symbol
def type_intersection_geometry_wrapper():
    return 'IG'


def name_geometry_wrapper():
    """ Selects the 'native' geometry wrapper of the kernel """
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")

    if it == 'cell':
        return name_element_geometry_wrapper()
    if it == 'exterior_facet' or it == 'interior_facet':
        return name_intersection_geometry_wrapper()
    assert False


def type_geometry_wrapper():
    """ Selects the 'native' geometry wrapper of the kernel """
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")

    if it == 'cell':
        return type_element_geometry_wrapper()
    if it == 'exterior_facet' or it == 'interior_facet':
        return type_intersection_geometry_wrapper()
    assert False


@preamble
def define_restricted_cell(name, restriction):
    ig = name_intersection_geometry_wrapper()
    which = "inside" if restriction == Restriction.INSIDE else "outside"
    return "const auto& {} = {}.{}();".format(name,
                                              ig,
                                              which,
                                              )


@symbol
def _name_cell(restriction):
    if restriction == Restriction.NONE:
        eg = name_element_geometry_wrapper()
        return "{}.entity()".format(eg)

    which = "inside" if restriction == Restriction.INSIDE else "outside"
    name = "cell_{}".format(which)
    define_restricted_cell(name, restriction)
    return name


def name_cell():
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")

    if it == 'cell':
        r = Restriction.NONE
    if it == 'exterior_facet':
        r = Restriction.INSIDE
    if it == 'interior_facet':
        raise NotImplementedError

    return _name_cell(r)


@preamble
def define_cell_geometry(name):
    eg = name_element_geometry_wrapper()
    return "auto {} = {}.geometry();".format(name,
                                             eg
                                             )


@symbol
def name_cell_geometry():
    define_cell_geometry("cell_geo")
    return "cell_geo"


@preamble
def define_intersection_geometry(name):
    ig = name_intersection_geometry_wrapper()
    return "auto {} = {}.geometry();".format(name,
                                             ig,
                                             )


@symbol
def name_intersection_geometry():
    define_intersection_geometry("is_geo")
    return "is_geo"


@symbol
def name_intersection():
    ig = name_intersection_geometry_wrapper()
    return "{}.intersection()".format(ig)


def name_geometry():
    """ Selects the 'native' geometry of the kernel """
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")

    if it == 'cell':
        return name_cell_geometry()
    if it == 'exterior_facet' or it == 'interior_facet':
        return name_intersection_geometry()
    assert False


@preamble
def define_in_cell_geometry(restriction, name):
    cell = _name_cell(restriction)
    ig = name_intersection_geometry_wrapper()
    which = "In" if restriction == Restriction.INSIDE else "Out"
    return "auto {} = {}.geometryIn{}side();".format(name,
                                                     ig,
                                                     which
                                                     )


@symbol
def name_in_cell_geometry(restriction):
    assert restriction is not Restriction.NONE

    name = "geo_in_{}side".format("in" if restriction is Restriction.INSIDE else "out")
    define_in_cell_geometry(restriction, name)
    return name


def apply_in_cell_transformation(name, local, restriction):
    geo = name_in_cell_geometry(restriction)
    return quadrature_preamble("{} = {}.global({});".format(name,
                                                            geo,
                                                            local,
                                                            ),
                               assignees=name,
                               )


@symbol
def name_in_cell_coordinates(local, basename, restriction):
    name = "{}_in_inside".format(basename)
    temporary_variable(name, shape=(name_dimension(),), shape_impl=("fv",))
    apply_in_cell_transformation(name, local, restriction=Restriction.INSIDE)
    return name


def to_cell_coordinates(local, basename):
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")

    if it == 'cell':
        return local
    if it == 'exterior_facet':
        return name_in_cell_coordinates(local, basename, Restriction.INSIDE)
    if it == 'interior_facet':
        raise NotImplementedError


@preamble
def define_dimension(name):
    geo = type_geometry_wrapper()
    return 'const int {} = {}::Entity::dimension;'.format(name, geo)


@symbol
def name_dimension():
    define_dimension("dim")
    return "dim"


@symbol
def type_jacobian_inverse_transposed():
    geo = type_element_geometry_wrapper()
    return "typename {}::Geometry::JacobianInverseTransposed".format(geo)


@preamble
def define_jacobian_inverse_transposed_temporary(name, shape, shape_impl):
    t = type_jacobian_inverse_transposed()
    return "{} {};".format(t,
                           name,
                           )


def define_jacobian_inverse_transposed(name):
    dim = name_dimension()
    temporary_variable(name, decl_method=define_jacobian_inverse_transposed_temporary, shape=(dim, dim))
    geo = name_cell_geometry()
    pos = name_quadrature_position()
    return quadrature_preamble("{} = {}.jacobianInverseTransposed({});".format(name,
                                                                               geo,
                                                                               pos,
                                                                               ),
                               assignees=frozenset({name}),
                               )


@symbol
def name_jacobian_inverse_transposed():
    define_jacobian_inverse_transposed("jit")
    return "jit"
